open Unix

type system_error =
| EIO       (* Erreur de bas niveau lors d'un accès au disque *)
| EACCESS   (* Erreur d'accès *)
| EISDIR    (* Répertoire trouvé où un fichier régulier est attendu *)
| ENOTDIR   (* Fichier régulier trouvé où un répertoire est attendu *)
| EINVAL    (* Argument invalide *)
| EEXIST    (* Un fichier existe déjà *)
| ENOENT    (* Un fichier manque *)
| EBUSY     (* Un fichier est occupé *)
| ENOSPC    (* Plus d'espace sur le disque *)

exception System of system_error * string * string

let system_error err call mes = raise (System (err, call, mes))

let handle_unix f x =
  try f x with Unix.Unix_error (error,s,mes) -> system_error EIO s
        (Format.sprintf "%s/%s" mes (Unix.error_message error))


(* --------------------- *)
(* Structures de données *)
(* --------------------- *)

(* Le type [partition] contient les principales valeurs caractérisant
   une partition. *)
type partition =
    { part_name: string;
      (* le nom du fichier sur le disque *)
      part_fd: Unix.file_descr;
      (* le descripteur de fichier correspondant *)
      block_size: int;
      (* la taille des blocs (en octects) *)
      block_nb: int;
      (* le nombre de blocs *)
      part_alive: bool;
      (* un booléen indiquant si la partition est activée, c'est a
         dire si le champ part_fd est toujours
         ouvert *)
      block_word_size: int;
      (* le nombre de mots par bloc (entiers de 4 octects dans un
         bloc) *)
      blocktbl_size: int;
      (* le nombre maximal de blocs dans un fichier, i.e. la taille en
         mots de la table des blocs d'un inode *)
      max_file_size: int;
      (* la taille maximale d'un fichier, calculee a partir des
         proprietes de la partition *)
      blocktbl_offset: int;
      (* la position de la table des blocs dans un inode, en
         octects.  *)
      inode_nb: int;
      (* le nombre d'inodes dans la partition *)
      last_free_inode: int;
      (* le numéro du dernier inode libéré *)
      free_block_list: int;
      (* le numéro du bloc index de la liste de blocs libres *)
      free_block_nb: int;
      (* la position dans ce bloc d'index *)
      fstype: string;
      (* le type de la partition *)
      root_inode: int;
      (* le numéro de l'inode du répertoire racine *)
    }

(* Définition d'un i-node *)
type inode =
    { stats: stats;
      blocktbl: int array;
      partition: partition; }

and file_kind = S_REG | S_DIR

and stats =
    { st_dev: int;
      st_ino: int;
      mutable st_kind: file_kind;
      mutable st_nlink: int;
      mutable st_size: int; }

(* caractéristique d'un répertoire *)
let dirent_size = 32
let filename_max_size = 28

(* --------------------- *)
(* Fonctions utilitaires *)
(* --------------------- *)

exception Implementation of string

let implementation_error s = raise (Implementation s)

let rec really_read fd buffer start length =
  if length <= 0 then () else
  match Unix.read fd buffer start length with
  | 0 -> raise End_of_file
  | r -> really_read fd buffer (start + r) (length - r);;

(* [read_int] permet de lire des entiers codés sur 4 caractères. *)
let read_int =
  let get_int8 s pos = int_of_char s.[pos] in
  fun s pos ->
    let c4 = get_int8 s pos in
    let c3 = get_int8 s (pos+1) in
    let c2 = get_int8 s (pos+2) in
    let c1 = get_int8 s (pos+3) in
    c1 + (c2 lsl 8) + (c3 lsl 16) + (c4 lsl 24)


let print_partition part =
  Format.printf "Partition %s (%s)@\n  @["
    part.part_name
    (if part.part_alive then "alive" else "dead");
  Format.printf "block_size: %i@\n" part.block_size;
  Format.printf "block_nb: %i@\n" part.block_nb;
  Format.printf "block_word_size: %i@\n" part.block_word_size;
  Format.printf "i-nodes@\n  @[";
  Format.printf "blocktbl_size: %i@\n" part.blocktbl_size;
  Format.printf "max_file_size: %i@\n" part.max_file_size;
  Format.printf "blocktbl_offset: %i@]@\n" part.blocktbl_offset;
  Format.printf "supernode@\n  @[";
  Format.printf "inode_nb: %i@\n" part.inode_nb;
  Format.printf "last_free_inode: %i@\n" part.last_free_inode;
  Format.printf "free_block_list: %i@\n" part.free_block_list;
  Format.printf "free_block_nb: %i@\n" part.free_block_nb;
  Format.printf "fstype: %s@\n" part.fstype;
  Format.printf "root_inode: %i@]@]@\n" part.root_inode

let print_block part bid b =
  Format.printf "Block %i:@\n  @[" bid;
  for i = 0 to (part.block_size / 4) - 1 do
    Format.printf "%i@ " (read_int b (i*4))
  done;
  Format.printf "@\n@?"


(* ------------------------- *)
(* Début du code à completer *)
(* ------------------------- *)

let open_partition name =
  (implementation_error "open_partition" : partition)

let lseek_block part n =
  (implementation_error "lseek_block" : unit)

let read_block part n =
  (implementation_error "read_block" : string)

let read_inode part n =
  (implementation_error "read_inode" : inode)


type file_descr = { inode : inode; mutable pos : int; mutable closed : bool }

let open_inode inode =
  { inode = inode;
    pos = 0;
    closed = false; }

let read desc buffer from len =
  let part = desc.inode.partition in
  if desc.closed then system_error EINVAL "read" "closed descr";
  if desc.inode.stats.st_size > part.max_file_size then
    system_error EIO "read" "Inconsistent file system";
  (implementation_error "read" : int)


let rec read_dirent desc =
  (implementation_error "read_dirent" : int * string)

let rec print_directory inode dirname =
  let p = inode.partition in
  let fd = open_inode inode in
  try
    while true do
      let (i, name) = read_dirent fd in
      let filename = Filename.concat dirname name in
      Format.printf "%i %s@\n" i filename;
      let inode = read_inode p i in
      match inode.stats.st_kind with
      | S_REG -> ()
      | S_DIR ->
          if name <> "." && name <> ".." then
            print_directory inode filename
    done
  with End_of_file -> ()


type block_type =
  | Inode of int    (* Inode, nombre de liens durs observés *)
  | Block of bool   (* Block, utilisé ou pas *)
  | IndexBlock      (* Block d'index de la liste des blocs libres *)
  | Supernode       (* Supernode du système *)
  | FreeBlock       (* Block libre confirmé *)

let string_of_block t =
  match t with
  | Inode n -> Format.sprintf "inode[%d]" n
  | Block b -> Format.sprintf "block[%b]" b
  | IndexBlock -> "indexblock"
  | Supernode -> "supernode"
  | FreeBlock -> "freeblock"

let fsck p =
  (* initialisation de la table des blocs *)
  let t = Array.create p.block_nb (Block false) in
  t.(0) <- Supernode;
  for i = 1 to p.inode_nb do
    t.(i) <- Inode 0
  done;

  (* parcours de répertoires *)
  Format.printf "Scanning directory tree@\n";
  let rec fsck_inode inode =
    Format.printf ".";
    match t.(inode) with
    | Inode 0 ->
        t.(inode) <- Inode 1;
        (implementation_error "fsck_inode" : unit)
    | Inode n ->
        implementation_error "fsck_inode"
    | _ ->
        implementation_error "fsck_inode"
  in
  fsck_inode p.root_inode;
  Format.printf "@\n";

  (* vérification des inodes *)
  for n = 1 to p.inode_nb do
    (implementation_error "Vérification du nombre de liens durs" : unit)
  done;

  (* parcours de la liste des blocs libres *)
  Format.printf "Scanning free block list@\n";
  (implementation_error "Scanning free block list" : unit);


  (* vérifier que tous les blocs ont été traités *)
  (implementation_error "Forgotten block" : unit);

  Format.printf "Scan done@\n@?";
  ()



let () =
  let usage =  Format.sprintf "Usage: %s file" Sys.argv.(0) in
  if not (Array.length Sys.argv = 2) then begin
    Format.eprintf "%s@." usage;
    exit 1
  end;
  let part = open_partition Sys.argv.(1) in
  print_partition part
